import json
import re
from typing import Any, Dict, Generic, Iterable, List, Optional, Tuple, TypeVar, Union
from uuid import UUID

from twisted.web.error import Error
from twisted.web.http import Request, BAD_REQUEST

try:
    from jsonschema import ValidationError
    from jsonschema.validators import validator_for
except ImportError:
    def validator_for(*args, **kwargs):  # type: ignore
        raise NotImplementedError("Install with jsonschema extra to enable validating JSON parameters")

    class ValidationError(Exception):  # type: ignore
        pass


T = TypeVar("T")


class BaseParam(Generic[T]):
    """
    Base class for retwist Parameters. Subclass and override parse() to implement custom parsing behavior.
    """

    def __init__(self, required=False, default=None, name=None):
        # type: (bool, Optional[T], Optional[str]) -> None
        """
        :param required: Throw 400 error if parameter is missing
        :param default: Default value, in case parameter missing
        :param name: Parameter name in the URL. Use this parameter if your parameter name is a reserved Python keyword
        """
        if required and default:
            raise ValueError("Required parameters can't have a default")

        self.required = required
        self.default = default
        self.name = name

    def parse_from_request(self, name, request):
        # type: (str, Request) -> Optional[T]
        """
        Parse parameter by name from request object. Throws 400 client error if parameter is required, but missing.
        :param name: Name of parameter in query
        :param request: Twisted request object
        :return: Parsed value
        """
        name_bytes = name.encode()
        if name_bytes not in request.args:
            if self.default is not None:
                return self.default
            if self.required:
                raise Error(BAD_REQUEST, message=b"Required")
            else:
                return None

        if len(request.args[name_bytes]) != 1:
            raise Error(BAD_REQUEST, message=b"Pass exactly one argument")

        val = request.args[name_bytes][0]
        return self.parse(val)

    def parse(self, val):
        # type: (bytes) -> T
        """
        Parse parameter from raw string from URL query. Override this for custom parsing behavior.
        :param val: Value as received in URL
        :return: Parsed value
        """
        raise NotImplementedError


class Param(BaseParam[str]):

    def parse(self, val):
        # type: (bytes) -> str
        """
        Parse parameter from raw string from URL query. Override this for custom parsing behavior.
        :param val: Value as received in URL
        :return: Parsed value
        """
        return val.decode()


class BoolParam(BaseParam[bool]):
    """
    Evaluates to True if "true" is passed, False otherwise.
    """

    def parse(self, val):
        # type: (bytes) -> bool
        val_str = val.decode()
        if val_str is None or val_str == "false":
            return False
        if val_str == "true":
            return True
        raise Error(BAD_REQUEST, message=b"Boolean parameter must be 'true' or 'false'")


class IntParam(BaseParam[int]):
    """
    Parse an integer.
    """

    def __init__(self, min_val=None, max_val=None, *args, **kwargs):
        # type: (Optional[int], Optional[int], *Any, **Any) -> None
        """
        :param min_val: Raise error if value is smaller than this
        :param max_val: Raise error if value is bigger than this
        """
        super(IntParam, self).__init__(*args, **kwargs)
        self.min_val = min_val
        self.max_val = max_val

    def parse(self, val):
        # type: (bytes) -> int
        val_str = val.decode()
        try:
            val_int = int(val_str)
        except (TypeError, ValueError):
            raise Error(BAD_REQUEST, b"Invalid integer: %s" % val)
        if self.min_val is not None and val_int < self.min_val:
            raise Error(BAD_REQUEST, b"Minimum value %d" % self.min_val)
        if self.max_val is not None and val_int > self.max_val:
            raise Error(BAD_REQUEST, b"Maximum value %d" % self.max_val)
        return val_int


class EnumParam(BaseParam[str]):
    """
    Allow a pre-defined list of string values; raise 400 client error otherwise.
    """

    def __init__(self, enum, *args, **kwargs):
        # type: (Iterable[str], *Any, **Any) -> None
        """
        :param enum: List/tuple of allowed values, e.g. ("enabled", "disabled")
        """
        super(EnumParam, self).__init__(*args, **kwargs)
        self.enum = frozenset(enum)

    def parse(self, val):
        # type: (bytes) -> str
        val_str = val.decode()
        if val_str not in self.enum:
            error_msg = b"Parameter must be one of %s" % str(sorted(self.enum)).encode()
            raise Error(BAD_REQUEST, error_msg)
        return val_str


class LangParam(Param):
    """
    Parse language from "lang" URL parameter, or infer from Accept-Language HTTP header.
    """

    accept_language_re = re.compile(r"([a-z]{1,8}(?:-[a-z]{1,8})?)\s*(?:;\s*q\s*=\s*(1|0\.[0-9]+))?", re.IGNORECASE)

    def parse_from_request(self, name, request):
        # type: (str, Request) -> Optional[str]
        if b"lang" in request.args:
            return super(LangParam, self).parse_from_request(name, request)
        return self.infer_lang(request)

    @classmethod
    def parse_accept_language(cls, accept_language_str):
        # type: (str) -> List[Tuple[str, str]]
        """
        Parse a list of accepted languages from an HTTP header.
        :param accept_language_str: HTTP Accept-Language value
        :return: A list of tuples (locale, weight), ordered by descending weight.
        """
        return sorted(
            (
                (match.group(1), match.group(2) or "1")
                for match in cls.accept_language_re.finditer(accept_language_str)
                if match.group(1) != "*"  # ignore wildcard language
            ),
            key=lambda lang_weight: float(lang_weight[1]), reverse=True
        )

    def infer_lang(self, request):
        # type: (Request) -> str
        http_header = request.getHeader("Accept-Language")
        default_lang = str(self.default)  # type: str
        if http_header:
            try:
                locales = self.parse_accept_language(http_header)
            except (TypeError, ValueError):
                return default_lang
            else:
                if locales:
                    return locales[0][0]
        return default_lang


class VersionParam(BaseParam[Tuple[int, ...]]):
    """
    Parse a version from a version string. Return a representation suitable for comparisons, e.g.:

    parse("5.9") < parse("5.10")
    """

    def parse(self, val):
        # type: (bytes) -> Tuple[int, ...]
        val_str = val.decode()
        try:
            return tuple(map(int, val_str.split(".")))
        except (TypeError, ValueError):
            raise Error(BAD_REQUEST, b"Invalid version literal")


# Approximate representation of JSON data, for type annotations
_JSON_OBJECT_TYPE = Dict[str, Any]
_JSON_TYPE = Union[None, bool, int, float, str, list, _JSON_OBJECT_TYPE]


class JsonParam(BaseParam[_JSON_TYPE]):
    """
    Parse a parameter encoded as JSON. You can optionally verify its data format by passing a https://json-schema.org/
    """

    # Some common JSON schema types. Can be passed to `JsonParam.array(items=JsonParam.NUMBER_TYPE)`, for example.
    NUMBER_TYPE = {"type": "number"}
    STRING_TYPE = {"type": "string"}
    UUID_TYPE = {"type": "string", "pattern": "^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$"}

    def __init__(self, schema=None, *args, **kwargs):
        # type: (Any, *Any, **Any) -> None
        """
        :param schema: JSON schema (draft 7 compatible), or None if format shouldn't be checked
        """
        super(JsonParam, self).__init__(*args, **kwargs)
        if schema is None:
            self.validator = None
        else:
            validator_class = validator_for(schema)
            validator_class.check_schema(schema)
            self.validator = validator_class(schema)

    def parse(self, val):
        # type: (bytes) -> _JSON_TYPE
        val_str = val.decode()
        try:
            data = json.loads(val_str)  # type: _JSON_TYPE
            if self.validator is not None:
                self.validator.validate(data)
        except ValueError as ex:
            error_message = str(ex).encode("utf-8")
            raise Error(BAD_REQUEST, b"Invalid JSON: %s" % error_message)
        except ValidationError as ex:
            error_message = ex.message.encode("utf-8")
            raise Error(BAD_REQUEST, b"JSON schema error: %s" % error_message)
        else:
            return data

    @classmethod
    def array(cls, items=None, min_items=None, max_items=None, *args, **kwargs):
        # type: (Optional[Dict[str, Any]], Optional[int], Optional[int], *Any, **Any) -> JsonParam
        """
        Parse a parameter encoded as a JSON array.

        :param items: Schema to apply for array items (optional)
        :param min_items: Minimum number of items (optional)
        :param max_items: Maximum number of items (optional)
        """
        schema = {"type": "array"}  # type: _JSON_OBJECT_TYPE
        if items is not None:
            schema["items"] = items
        if min_items is not None:
            schema["minItems"] = min_items
        if max_items is not None:
            schema["maxItems"] = max_items
        return cls(schema, *args, **kwargs)


class UUIDParam(BaseParam[UUID]):
    """
    Parameter that verifies it's a valid UUID.
    """

    MALFORMED_ERROR_MSG = b"Malformed UUID"

    def parse(self, val):
        # type: (bytes) -> UUID
        val_str = val.decode()
        try:
            return UUID(val_str)
        except ValueError:
            raise Error(BAD_REQUEST, message=UUIDParam.MALFORMED_ERROR_MSG)



from logging import getLogger
from typing import Any, Set

from twisted.internet.defer import Deferred, DeferredList
from twisted.internet.interfaces import IListeningPort, IReactorCore
from twisted.web.http import Request
from twisted.web.server import Site

logger = getLogger(__name__)


def wait_on_shutdown(reactor, site, port, timeout):
    # type: (IReactorCore, Site, IListeningPort, float) -> None
    """
    Change a running site to wait for pending requests on shutdown.

    When the reactor is to be shut down (e.g. on SIGINT), incoming connections are no longer accepted, but pending
    requests are allowed to execute until a timeout is hit.

    :param reactor: Reactor that is running your site, likely twisted.internet.reactor
    :param site: twisted.web.server.Site instance which is serving requests
    :param port: IListeningPort instance, e.g. as returned by reactor.listenTCP
    :param timeout: Timeout in seconds after which pending requests are canceled and shutdown is forced
    """
    running_reqs = set()  # type: Set[int]
    wrapped_request_factory = site.requestFactory

    def request_factory(*args, **kwargs):
        # type: (*Any, **Any) -> Request
        """
        Request factor for a Twisted connection factory. During their execution, requests are kept in a set.
        """
        def remove_req(_, req_hash):
            # type: (Any, int) -> None
            running_reqs.remove(req_hash)

        req = wrapped_request_factory(*args, **kwargs)

        req_hash = id(req)
        running_reqs.add(req_hash)
        req.notifyFinish().addBoth(remove_req, req_hash)

        return req

    def shutdown():
        # type: () -> Deferred
        """
        Shutdown handler for the Twisted reactor. Stop accepting incoming requests. Then, delay the shutdown of the
        reactor up to a certain timeout if there are still requests running.
        """
        logger.info("Shutdown requested")

        # force shutdown after timeout
        timeout_deferred = Deferred()

        def kill_timeout():
            # type: () -> None
            logger.warning("Timeout reached, canceling %s requests", len(running_reqs))
            timeout_deferred.callback(True)
        reactor.callLater(timeout, kill_timeout)

        # stop listening to new requests
        close_conn_deferred = port.stopListening()

        def kill_if_requests_done():
            # type: () -> None
            """
            Call the timeout if all requests are done; otherwise wait a few more seconds.
            """
            if not running_reqs:
                logger.info("No pending requests, shutting down!")
                timeout_deferred.callback(True)
            else:
                logger.info("%s requests still running, waiting ...", len(running_reqs))
                reactor.callLater(5, kill_if_requests_done)
        kill_if_requests_done()

        return DeferredList([timeout_deferred, close_conn_deferred])

    site.requestFactory = request_factory
    reactor.addSystemEventTrigger("before", "shutdown", shutdown)



from typing import Any, Dict

import sentry_sdk
from twisted.python import log
from twisted.python.failure import Failure


def add_request_context_to_scope(context, scope):
    # type: (Dict[str, Any], sentry_sdk.Scope) -> None
    """
    Add context information which is useful for debugging from a Twisted request to a Sentry scope.
    """
    if "user_id" in context:
        scope.set_user({"id": context.pop("user_id", None)})

    scope.set_context("request", context)


def log_to_sentry(event):
    # type: (Dict[str, Any]) -> None
    """
    Twisted log observer for reporting errors to sentry.
    :param event: Twisted log event dictionary.
    """
    if not event.get("isError") or "failure" not in event:
        return

    failure = event["failure"]  # type: Failure
    exc = failure.value
    exc_tuple = (type(exc), exc, failure.getTracebackObject())

    if "context" in event:
        # If a Twisted request has been added as context to the logged event, we can extract useful debug info
        context = event["context"]  # type: Dict[str, Any]
        with sentry_sdk.push_scope() as scope:
            add_request_context_to_scope(context, scope)
            sentry_sdk.capture_exception(exc_tuple)
    else:
        sentry_sdk.capture_exception(exc_tuple)


def enable_sentry_reporting():
    # type: () -> None
    """
    Enable Sentry logging for any errors reported via twisted.python.log.err.

    Call sentry_sdk.init() before this.
    """
    log.addObserver(log_to_sentry)



# flake8: noqa
from retwist.param import Param
from retwist.param import BoolParam, EnumParam, IntParam, JsonParam, LangParam, UUIDParam, VersionParam
from retwist.param_resource import ParamResource
from retwist.json_resource import JsonResource
from retwist.route_site import RouteSite



import json
import re
from logging import getLogger
from typing import Any, Dict, Optional, Union

try:
    from inspect import iscoroutinefunction
except ImportError:
    # We must be on Python < 3.5!
    def iscoroutinefunction(func):  # type: ignore
        return False

from twisted.internet.defer import CancelledError, Deferred, ensureDeferred, maybeDeferred
from twisted.internet.error import ConnectError, ConnectionDone, ConnectionLost
from twisted.python import log
from twisted.python.failure import Failure
from twisted.web.error import Error
from twisted.web.resource import Resource
from twisted.web.http import Request, BAD_REQUEST, INTERNAL_SERVER_ERROR
from twisted.web.server import NOT_DONE_YET

from retwist.param_resource import ParamResource

logger = getLogger(__name__)


class JsonResource(ParamResource):
    """
    Twisted resource with convenience methods to encode JSON responses. Supports JSONP.
    """

    encoding = "utf-8"
    jsonp_callback_re = re.compile(br"^[_a-zA-Z0-9\.$]+$")

    @classmethod
    def json_dump_default(cls, o):
        # type: (Any) -> Any
        """
        Override this to implement custom JSON encoding. Passed to json.dump method.
        :param o: Object which cannot be JSON-serialized.
        :return: JSON-serializable object (or throw TypeError)
        """
        raise TypeError("Can't JSON serialize {}".format(type(o).__name__))

    def json_GET(self, request):
        # type: (Request) -> Any
        """
        Override this to return JSON data to render.
        :param request: Twisted request.
        """
        raise NotImplementedError()

    def render(self, request):
        # type: (Request) -> Union[Any, bytes]
        """
        Before we render this request as normal, parse parameters, and add them to the request! Also, catch any errors
        during parameter parsing, and show them appropriately.
        :param request: Twisted request object
        :return: Byte string or NOT_DONE_YET - see IResource.render
        """
        try:
            request.url_args = self.parse_args(request)
        except Exception:
            self.handle_exception(request)
            return NOT_DONE_YET
        else:
            return Resource.render(self, request)

    def render_GET(self, request):
        # type: (Request) -> Any
        """
        Get JSON data from json_GET, and render for the client.

        Do not override in sub classes ...
        :param request: Twisted request
        """
        if iscoroutinefunction(self.json_GET):
            coroutine = self.json_GET(request)
            json_def = ensureDeferred(coroutine)  # type: Deferred
        else:
            json_def = maybeDeferred(self.json_GET, request)

        json_def.addCallback(self.send_json_response, request)
        json_def.addErrback(self.handle_failure, request)

        # handle connection failures
        request.notifyFinish().addErrback(self.on_connection_closed, json_def)

        return NOT_DONE_YET

    def response_envelope(self, response, status_code=200, status_message=None):
        # type: (Any, int, Optional[str]) -> Any
        """
        Implement this to transform JSON responses before sending, e.g. by putting HTTP status codes in the response.
        :param response: JSON data about to be sent to client
        :param status_code: HTTP status code
        :param status_message: Optional status message, e.g. error message
        :return: Wrapped JSON-serializable data
        """
        return response

    def send_json_response(self, response, request, status_code=200, status_message=None):
        # type: (Any, Request, int, Optional[str]) -> None
        """
        Send JSON data to client.
        :param response: JSON-serializable data
        :param request: Twisted request
        :param status_code: HTTP status code
        :param status_message: Optional status message, e.g. error message
        """
        is_jsonp = len(request.args.get(b"callback", [])) == 1
        if is_jsonp:
            callback = request.args[b"callback"][0]
            if not self.jsonp_callback_re.match(callback):
                del request.args[b"callback"]
                return self.send_json_response("Invalid callback", request, status_code=BAD_REQUEST)
            request.setHeader(b"Content-Type", b"application/javascript; charset=%s" % self.encoding.encode())
            request.write(callback + b"(")
        else:
            request.setHeader(b"Content-Type", b"application/json; charset=%s" % self.encoding.encode())
            request.setResponseCode(status_code)

        response = self.response_envelope(response, status_code=status_code, status_message=status_message)
        response_str = json.dumps(response, allow_nan=False, ensure_ascii=False, check_circular=False,
                                  default=self.json_dump_default)
        response_bytes = response_str.encode(self.encoding)
        request.write(response_bytes)

        if is_jsonp:
            request.write(b")")

        request.finish()

    def handle_exception(self, request):
        # type: (Request) -> None
        """
        Handle the most recent exception.
        :param request: Twisted request
        """
        self.handle_failure(Failure(), request)

    def handle_failure(self, failure, request):
        # type: (Failure, Request) -> None
        """
        Handle a failure. For connection errors, we do nothing - no chance to send anything. For client errors, we show
        an informative error message. For server errors, we show a generic message, and log the error.
        :param failure: Failure instance
        :param request: Twisted request
        """
        exception = failure.value  # type: Exception

        # Connection errors. This is business as usual on the Internet. We do nothing - we can't reach the client to
        # tell them about it anyways, and it's not worth logging.
        if any(isinstance(exception, exc_type) for exc_type in {
            CancelledError, ConnectError, ConnectionDone, ConnectionLost
        }):
            return

        # Client error - we expose the error message to the client, but don't log anything.
        if isinstance(exception, Error):
            web_error = exception  # type: Error
            status_code = int(web_error.status)
            if 400 <= status_code < 500:
                message = str(exception)
                self.send_error(status_code, message, request)
                return

        # Unhandled server error - let's log it
        log.err(failure, context=self.get_context(request))

        # We don't let the client see any part of the exception, since it might expose internals
        self.send_error(INTERNAL_SERVER_ERROR, "Server-side error", request)

    def send_error(self, status_code, message, request):
        # type: (int, str, Request) -> None
        """
        Send error message to the client.
        :param status_code: HTTP error code
        :param message: Error message that we want to expose to the client
        :param request: Twisted request
        """
        self.send_json_response(None, request, status_code=status_code, status_message=message)

    def on_connection_closed(self, failure, deferred):
        # type: (Failure, Deferred) -> None
        """
        Handle connection errors.
        :param failure: Twisted failure
        :param deferred: The async call to json_GET
        """
        deferred.cancel()

    def get_context(self, request):
        # type: (Request) -> Dict[str, Any]
        """
        Override to add custom context.
        Override this method to add other custom fields.
        :param request: The request
        :return: Dict with the context
        """
        return {
            "url": request.uri,
            "method": request.method,
            "headers": request.getAllHeaders(),
            "query_string": request.uri.replace(request.path, b"").strip(b"?"),
            "data": request.args
        }



import re
from typing import Any, Dict, Pattern

from twisted.web.http import Request
from twisted.web.resource import NoResource, Resource
from twisted.web.server import Site


class RouteSite(Site):
    """
    Twisted site which allows looking up resources by routes. Routes are regular expressions which match the entire
    request path, e.g. "/hotels/1234/info".

    Routes can have placeholders in them, which are stored as arguments in request.path_args before your render_* sees
    it.
    """

    def __init__(self, resource=None, *args, **kwargs):
        # type: (Resource, *Any, **Any) -> None
        """
        :param resource: Root resource for Twisted's standard Site implementation. Pass a resource if you want to fall
        back to Twisted's default resource lookup mechanism in case no route matches. If None is passed, defaults to a
        NoResource() instance.
        """
        resource = resource or NoResource()
        Site.__init__(self, resource, *args, **kwargs)
        self.routes = {}  # type: Dict[Pattern[str], Resource]

    def addRoute(self, route, resource):
        # type: (str, Resource) -> None
        """
        Register a handler resource for a path.
        :param route: Regular expression string. If a request's path matches this regex, this resource's render_*
        method will be called with it. The request will have arguments parsed from the route added as request.path_args.
        If there are named groups in the regular expression, the path arguments are passed as a dictionary. If they're
        unnamed, they're passed as a tuple.
        :param resource: A resource instance. Its render_* method will be called for requests matching route pattern.
        """
        route_re = re.compile(route)
        self.routes[route_re] = resource

    def getResourceFor(self, request):
        # type: (Request) -> Resource
        """
        Check if a route matches this request. Fall back to Twisted default lookup behavior otherwise.
        :param request: Twisted request instance
        :return: Resource to handle this request
        """
        request.site = self

        for route_re, resource in self.routes.items():
            path = request.path.decode()
            match = route_re.match(path)
            if match:
                request.path_args = match.groupdict() or match.groups()
                return resource

        return Site.getResourceFor(self, request)



import collections
from typing import Any, Callable, Iterable

from twisted.internet.defer import Deferred, FirstError
from twisted.python.failure import Failure


class LimitedDeferredList(Deferred):
    """
    Run a list of Deferreds in parallel, much like DeferredList, but limit the number of concurrently executed
    Deferreds.

    Useful if you need to run 100 things in parallel, but don't want to run more than 10 at once, e.g. to limit access
    to system resources.

    Unlike DeferredList, you can't pass a list of Deferreds to this class, since they would be running already. Instead,
    pass a list of factory methods that make the Deferred on demand.

    If any single one of the Deferreds fails, the overall LimitedDeferredList will errback with a FirstError failure,
    just like the DeferredList.
    """

    def __init__(self, deferred_factories, max_concurrent):
        # type: (Iterable[Callable[[], Deferred]], int) -> None
        """
        :param deferred_factories: A list of callables which, when invoked without arguments, return a Deferred.
        :param max_concurrent: Run no more than max_concurrent Deferreds at the same time.
        """
        Deferred.__init__(self)

        # Index the Deferreds so we can order the results when they callback
        indexed_factories = enumerate(deferred_factories)
        self.deferred_factories = collections.deque(indexed_factories)

        self.results = [None] * len(self.deferred_factories)
        self.counter = 0

        # Start up max_concurrent Deferreds! Later, we start a new Deferred for every one that callbacks. So we don't
        # even need to store max_concurrent
        for _ in range(max_concurrent):
            self.__schedule_deferred()

    def __deferred_callback(self, result, index):
        # type: (Any, int) -> None
        """
        Callback for every one of the scheduled Deferreds.
        :param result: Result of the Deferred.
        :param index: Index in overall result list.
        """
        self.counter -= 1
        self.results[index] = result
        # If there are no other Deferreds running and no more to be scheduled, we're all done. Resolve with the results
        # list
        if not self.__schedule_deferred() and self.counter == 0:
            self.callback(self.results)

    def __deferred_errback(self, failure, index):
        # type: (Failure, int) -> None
        """
        Errback for every one of the scheduled Deferreds. Make the whole LimitedDeferredList fail.
        :param failure: Failure object.
        :param index: Index in overall result list.
        """
        if not self.called:
            self.errback(FirstError(failure, index))

    def __schedule_deferred(self):
        # type: () -> bool
        """
        Schedule the next available Deferred.
        :return: True if a Deferred was scheduled, False if not (i.e. we're done with our work)
        """
        if self.deferred_factories:
            self.counter += 1
            index, deferred_factory = self.deferred_factories.pop()
            deferred = deferred_factory()  # type: Deferred
            deferred.addCallback(self.__deferred_callback, index)
            deferred.addErrback(self.__deferred_errback, index)
            return True
        return False



import logging
import os
from typing import Optional, Union

from twisted.python import log as twisted_log


def redirect_twisted_logging(logger_name=None, log_level=None):
    # type: (Optional[str], Optional[Union[int, str]]) -> None
    """
    Redirect Twisted's log messages to the Python logging module.

    Twisted has its own logging system (several, actually). Calling this method redirects all Twisted log messages to
    Python logging, by default as logger name "twisted", and for levels WARNING and up.

    :param logger_name: Name of the Twisted logger, so you can retrieve and configure it with `getLogger(...)`
    :param log_level: Log level for the messages to be forward. Default is WARNING
    """
    if logger_name is None:
        logger_name = "twisted"
    if log_level is None:
        log_level = logging.WARNING

    # Redirect messages from Twisted's log system to Python logging
    observer = twisted_log.PythonLoggingObserver(loggerName=logger_name)
    observer.start()

    # Adjust the log level, because debug and info messages are too verbose for most use cases
    logging.getLogger(logger_name).setLevel(log_level)

    # Twisted will by default print error messages to stderr until it has been instructed via `startLogging` to use a
    # different log handler. We silence Twisted's logging here, since we already redirect to Python's logging system.
    # There must be a better way to do this!
    twisted_log.startLogging(open(os.devnull, "w"))



from twisted.web.error import Error
from twisted.web.http import NOT_ALLOWED
from twisted.web.static import File


class NoListingFile(File):
    """
    Serve files, but disallow directory listing.
    """

    def directoryListing(self):
        # type: () -> None
        raise Error(NOT_ALLOWED, b"Not allowed")



import inspect
from typing import Any, Dict

from twisted.web.error import Error
from twisted.web.http import Request
from twisted.web.resource import Resource

from retwist.param import BaseParam


class ParamResource(Resource):
    """
    Twisted resource with convenient parsing of parameters.

    Parameters are defined at class level:

    age = retwist.Param()

    You can then retrieve parameters by calling parse_args(request) in your render_* method.
    """

    ERROR_MSG = b"Error in parameter %s: %s"

    def parse_args(self, request):
        # type: (Request) -> Dict[str, Any]
        """
        Parse arguments from request. Throws twisted.web.error.Error instances on client errors.

        :param request: Twisted request
        :return: Dictionary of parameter names to parsed values
        """
        args = {}
        for name, param in inspect.getmembers(self, lambda member: isinstance(member, BaseParam)):

            # Name can be overridden. This is useful if parameter name is a Python reserved keyword
            if param.name is not None:
                name = param.name

            try:
                val = param.parse_from_request(name, request)
            except Error as ex:
                error_msg = ParamResource.ERROR_MSG % (name.encode(), ex.message)
                raise Error(ex.status, error_msg)
            else:
                args[name] = val
        return args



import io
import os.path

import setuptools

version = "0.4.1"

this_directory = os.path.abspath(os.path.dirname(__file__))
readme_path = os.path.join(this_directory, "README.md")
with io.open(readme_path, encoding="utf-8") as f:
    long_description = f.read()

setuptools.setup(
    name="retwist",
    packages=setuptools.find_packages(exclude=["tests", "tests.*"]),
    package_data={
        "retwist": ["py.typed"],
    },
    description="Write JSON REST APIs in the Twisted framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="TrustYou",
    author_email="development@trustyou.com",
    version=version,
    download_url="https://github.com/trustyou/retwist/archive/v{}.tar.gz".format(version),
    url="https://github.com/trustyou/retwist",
    install_requires=[
        "twisted>16.4",
        "typing;python_version<'3.5'",
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    extras_require={
        "sentry": ["sentry-sdk"],
        "jsonschema": [
            "jsonschema",
            # pyrsistent is a dependency of jsonschema, pinned to the latest Python 2-compatible version
            "pyrsistent<=0.16.1;python_version<'3'"
        ]
    }
)
